/*
 * Copyright (C) 2018 Kalyst <contact@kalyst.fr>
 * Based on work by Wim Taymans <wim.taymans@gmail.com>, see:
 *   gst-plugins-good/tests/examples/rtp/client-PCMA.c
 *   gst-plugins-good/tests/examples/rtp/server-alsasrc-PCMA.c
 */

#include "psa.h"
#include "gst_media.h"

#define GLIB_DISABLE_DEPRECATION_WARNINGS

static GstElement *send_pipeline;
static GstElement *send_rtpbin;
static GMainLoop  *talker_loop;

static GstElement *recv_pipeline;
static GstElement *recv_rtpbin;
static GMainLoop  *listener_loop;

/*
 * This function is not available before 1.14 (see gstutils)
 */
const gchar *psa_state_change_get_name (GstStateChange transition)
{
  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      return "NULL->READY";
    case GST_STATE_CHANGE_READY_TO_PAUSED:
      return "READY->PAUSED";
    case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
      return "PAUSED->PLAYING";
    case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
      return "PLAYING->PAUSED";
    case GST_STATE_CHANGE_PAUSED_TO_READY:
      return "PAUSED->READY";
    case GST_STATE_CHANGE_READY_TO_NULL:
      return "READY->NULL";
    default:
      return "Unknown state return";
  }
}

GstStateChangeReturn psa_set_state_change(endpoint_t who,
                                          GstStateChange transition)
{
        GstElement *pipeline;
        GstStateChangeReturn ret;
        gchar *who_name;

        switch (who) {
        case LISTENER:
                pipeline = recv_pipeline;
                who_name = "listener";
                break;

        case TALKER:
                pipeline = send_pipeline;
                who_name = "talker";
                break;
                
        default:
                return GST_STATE_CHANGE_FAILURE;
        }
        
        ret = gst_element_change_state(pipeline, transition);
        g_print("%s: transition %s result \"%s\"\n", who_name,
                psa_state_change_get_name(transition),
                gst_element_state_get_name(ret));
        return ret;
}

void psa_gst_eos_event(endpoint_t who)
{
	if (who == LISTENER)
        	gst_element_send_event(recv_pipeline, gst_event_new_eos());
	else /* TALKER */
		gst_element_send_event(send_pipeline, gst_event_new_eos());
}

static gboolean bus_call_cb(G_GNUC_UNUSED GstBus *bus, GstMessage *msg,
                            gpointer data)
{
        GMainLoop *loop = (GMainLoop *) data;

        switch (GST_MESSAGE_TYPE(msg)) {

        case GST_MESSAGE_EOS:
                g_print ("End of stream\n");
                g_main_loop_quit (loop);
                break;

        case GST_MESSAGE_ERROR: {
                gchar  *debug;
                GError *error;
                gst_message_parse_error (msg, &error, &debug);
                g_free (debug);
                g_printerr ("Error: %s\n", error->message);
                g_error_free (error);
                g_main_loop_quit (loop);
                break;
        }
        case GST_MESSAGE_STATE_CHANGED: {
                GstState old_state, new_state;
                gchar *name;

                gst_message_parse_state_changed(msg, &old_state,
                                                &new_state, NULL);
                name = GST_OBJECT_NAME(msg->src);
                if (g_strcmp0(name, "psa-lab") == 0) {
                        g_print("Element %s changed state from %s to %s.\n",
                                GST_OBJECT_NAME (msg->src),
                                gst_element_state_get_name (old_state),
                                gst_element_state_get_name (new_state));
                }
                break;
        }
                
        default:
                break;
        }

        return TRUE;
}

static void rtcp_print_packet(gchar *who, guint8 *ptr)
{
        guint i;
        guint16 len;
        guint32 ntp_ts_h, cumul = 0;
        guint64 ntp_ts;
	rtcp_header_t *rtcp;
	GstRTCPType type;
        
        rtcp = (rtcp_header_t *) ptr;
        len = GUINT16_FROM_BE(rtcp->len)+1;
        type = rtcp->type;
        g_print("\n%s header(ver:%hhu,p:%hhu,rbc:%hhu,"
                "typ:%hhu,len:%hu,", who, rtcp->version,
                rtcp->padding, rtcp->reports, type, len);
        ptr += 4;
        g_print("snd:%u)", GST_READ_UINT32_BE(ptr));
        ptr += 4;
        if (type == GST_RTCP_TYPE_SR ||
            type == GST_RTCP_TYPE_RR)    { /* 200|201 */
                if (type == GST_RTCP_TYPE_SR) {
                        ntp_ts = GST_READ_UINT64_BE(ptr);
                        ntp_ts_h = (ntp_ts >> 32);
                        g_print("\n       ts(ntp:%u,", ntp_ts_h);
                        ptr += 8;
                        g_print("rtp:%u) ", GST_READ_UINT32_BE(ptr));
                        ptr += 4;
                        g_print("count(pkt:%u", GST_READ_UINT32_BE(ptr));
                        ptr += 4;
                        g_print(",byte:%u)", GST_READ_UINT32_BE(ptr));
                        ptr += 4;
                }
                
                /* affichage "Report Block"s éventuels */
                if (rtcp->reports == 0) {
                        g_print(" rbc(no)\n");
                }
                else {
                        g_print("\n       rb(");
                        for (i=0; i<rtcp->reports; i++) {
                                g_print("snd:%u,", GST_READ_UINT32_BE(ptr));
                                ptr += 4;
                                g_print("frl:%hhu", *ptr++);
                                cumul = *ptr++ << 16;
                                cumul |= GST_READ_UINT16_BE(ptr);
                                g_print(",cum:%u,", cumul);
                                ptr += 2;
                                g_print("hsq:%u,", GST_READ_UINT32_BE(ptr));
                                ptr += 4;
                                g_print("jit:%u,", GST_READ_UINT32_BE(ptr));
                                ptr += 4;
                                g_print("lsr:%u,", GST_READ_UINT32_BE(ptr));
                                ptr += 4;
                                g_print("dsr:%u", GST_READ_UINT32_BE(ptr));
                          }
                        g_print(")\n");
                }
        }
        
        else {
                /* TODO:
                 *   GST_RTCP_TYPE_BYE (to report EOS?)
                 *   GST_RTCP_TYPE_AVB (or using GST_RTCP_TYPE_APP?)
                 */
        }
}

G_GNUC_UNUSED
static GstPadProbeReturn rtcp_probe_cb (G_GNUC_UNUSED GstPad *pad,
					GstPadProbeInfo *info,
					G_GNUC_UNUSED gpointer user_data)
{
        GstMapInfo map;
        guint8 *ptr;
        GstBuffer *buffer;

        buffer = GST_PAD_PROBE_INFO_BUFFER(info);
        if (buffer == NULL)
                return GST_PAD_PROBE_OK;

       if (gst_buffer_map(buffer, &map, GST_MAP_READ)) {
                ptr = (guint8 *) map.data;
                rtcp_print_packet("probe:", ptr);
		gst_buffer_unmap(buffer, &map);
        }
        return GST_PAD_PROBE_OK;
}

/*
 * See rtpsession.c:282, on-sending-rtcp signal
 * This callback is called before sending an RTCP packet, it can be used
 * to add extra RTCP GST_RTCP_TYPE_APP Packets (ie at the end).
 * 
 * sessid: the object which received the signal
 * buffer: the GstBuffer containing the RTCP packet about to be sent
 * early:TRUE if the packet is early, %FALSE if it is regular
 *
 * return TRUE if the RTCP buffer should NOT be suppressed
 *        FALSE if suppressing is acceptable
 */

static gboolean on_sending_rtcp_cb(guint sessid, GstBuffer *buffer,
                                   gboolean early)
{
        GstRTCPBuffer rtcp = GST_RTCP_BUFFER_INIT;
        GstRTCPPacket packet;
        GstRTCPType code;
        gboolean cont;
        guint8 *data;

        /* AVB data fields... */
        guint32 ssrc;
	guint16 gm_time_base_indicator;	 // champ SHM PTP 
        guint16 gm_identity1;		 // 0 (not used) 
        guint64 gm_identity2;		 // champ gptp_grandmaster_id shm 
        guint64 stream_id;		 // à positionner au démarrage de gst 
        guint32 as_timestamp;            // partie médiane du system_time SHM
        guint32 rtp_timestamp;		 // duplication du RTP

        g_print("\n>on_sending_rtcp: (sess:%u,early:%d,size:%ld)\n", sessid,
                early, gst_buffer_get_size(buffer));
        if (!gst_rtcp_buffer_validate(buffer))
                goto end;

        if (!gst_rtcp_buffer_map(buffer, GST_MAP_READWRITE, &rtcp))
                goto end;
        
        g_print("> #packets: %d\n", gst_rtcp_buffer_get_packet_count(&rtcp));
        
        cont = gst_rtcp_buffer_get_first_packet(&rtcp, &packet);
        while (cont) {
                code = gst_rtcp_packet_get_type(&packet);
               if (code == GST_RTCP_TYPE_SR) {
                        gst_rtcp_packet_sr_get_sender_info(&packet, &ssrc,
                                                           NULL, &rtp_timestamp,
                                                           NULL, NULL);
                        g_print(">  packet: type %d, SR ssrc: %u, rtp ts: %u\n",
                                code, ssrc, rtp_timestamp);
                }
                else if (code == GST_RTCP_TYPE_RR) {
                        ssrc = gst_rtcp_packet_rr_get_ssrc(&packet);
                        g_print(">  packet: type %d, RR ssrc: %u\n",
                                code, ssrc);
                }
                else if (code == GST_RTCP_TYPE_BYE) {
                        g_print(">  packet: type %d, BYE ssrc:%u\n", code,
                                gst_rtcp_packet_bye_get_nth_ssrc(&packet, 0));
                        psa_gst_eos_event(LISTENER);
                        return FALSE;
                }
                else {
                        g_print(">  packet: type %d\n", code);
                }
               cont = gst_rtcp_packet_move_to_next(&packet);
        }

        if (!gst_rtcp_buffer_add_packet(&rtcp, GST_RTCP_TYPE_APP, &packet)) {
                g_print("> **** isbuf %s map %s isvalid %s\n",
                        GST_IS_BUFFER(rtcp.buffer)? "true": "false",
                        ((GST_RTCP_TYPE_APP != GST_RTCP_TYPE_INVALID)?
                         "true": "false"),
                        (rtcp.map.flags & GST_MAP_WRITE)? "true": "false");
        
                goto end;
        }
        g_print("> packet added: #packets: %d\n",
                gst_rtcp_buffer_get_packet_count(&rtcp));
        
	/* waiting 1.10+ gtreamer version */
        gst_rtcp_packet_app_set_subtype(&packet, 8);
        gst_rtcp_packet_app_set_ssrc(&packet, ssrc);
        gst_rtcp_packet_app_set_name(&packet, "PSA");
        if (!gst_rtcp_packet_app_set_data_length (&packet, 8)) {
                g_print("> set_data_length problem\n");
                gst_rtcp_packet_remove(&packet);
                return TRUE;
        }

        data = gst_rtcp_packet_app_get_data(&packet);
        
        gm_time_base_indicator = 0;
        GST_WRITE_UINT16_BE(data, gm_time_base_indicator);
        data += 2;
        gm_identity1 = 0;
        GST_WRITE_UINT16_BE(data, gm_identity1);
        data += 2;
        gm_identity2 = 0;
        GST_WRITE_UINT64_BE(data, gm_identity2);
        data += 8;
        stream_id = 0;
        GST_WRITE_UINT64_BE(data, stream_id);
        data += 8;
        as_timestamp = 0;
        GST_WRITE_UINT32_BE(data, as_timestamp);
        data += 4;
        GST_WRITE_UINT32_BE(data, rtp_timestamp);        

        gst_rtcp_buffer_unmap(&rtcp);
        g_print("> buffer unmap...\n");

end:
        return TRUE;
}

/*
 * See rtpsession.c:376, on_receiving_rtcp signal
 * This callback is called when receiving an RTCP packet before it
 * is handled by the session. It can be used to extract custom information
 * from RTCP packets.
 * 
 * sessid: the object which received the signal
 * buffer: the GstBuffer containing the RTCP packet that was received
 *
 * return void
 */
static void on_receiving_rtcp_cb(guint sessid, GstBuffer *buffer)
{
        GstRTCPBuffer rtcp = GST_RTCP_BUFFER_INIT;
        GstRTCPPacket packet;
        GstRTCPType code;
        gboolean cont;

        /* AVB data fields... */
        guint32 ssrc;
        /*
        guint16 gm_time_base_indicator;	champ SHM PTP
        guint16 gm_identity1;		0 (not used)
        guint64 gm_identity2;		champ gptp_grandmaster_id shm 
        guint64 stream_id;		à positionner au démarrage de gst
        guint32 as_timestamp;        partie médiane du system_time SHM */
        guint32 rtp_timestamp;		/* duplication du RTP */

        g_print("\n<on_receiving_rtcp: (sess:%u,size:%ld)\n", sessid,
                gst_buffer_get_size(buffer));

        if (!gst_rtcp_buffer_validate(buffer))
                goto end;

        if (!gst_rtcp_buffer_map(buffer, GST_MAP_READ, &rtcp))
                goto end;

        g_print("< #packets: %d\n", gst_rtcp_buffer_get_packet_count(&rtcp));
        
        cont = gst_rtcp_buffer_get_first_packet(&rtcp, &packet);
        while (cont) {
                code = gst_rtcp_packet_get_type(&packet);
                if (code == GST_RTCP_TYPE_SR) {
                        gst_rtcp_packet_sr_get_sender_info(&packet, &ssrc,
                                                           NULL, &rtp_timestamp,
                                                           NULL, NULL);
                        g_print("<  packet: type %d, SR ssrc: %u, rtp ts: %u\n",
                                code, ssrc, rtp_timestamp);
                }
                else if (code == GST_RTCP_TYPE_RR) {
                        g_print("<  packet: type %d, RR ssrc: %u\n", code,
                                gst_rtcp_packet_rr_get_ssrc(&packet));
                }
                else if (code == GST_RTCP_TYPE_BYE) {
                        g_print(">  packet: type %d, BYE ssrc:%u\n", code,
                                gst_rtcp_packet_bye_get_nth_ssrc(&packet, 0));
                        psa_gst_eos_event(LISTENER);
                        return;
                }
                else {
                        g_print("<  packet: type %d\n", code);
                }
                cont = gst_rtcp_packet_move_to_next(&packet);
        }

        /*
         * for future use (reminder)
         * guint8* gst_rtcp_packet_app_get_data(GstRTCPPacket * packet);
         */
        g_print("< **** packets to process... soon ****\n");

end:
        return;
}

/*
 *
 * talker side
 *
 *
 */

/*
 * callback pour associer decodebin et audioconvert côté talker
 */
static void talker_new_pad_cb(GstElement *element, G_GNUC_UNUSED GstPad *pad,
                              gpointer data)
{
        gst_element_link(element, (GstElement *) data);
}

/*
 * psa_gst_send_init builds a pipeline equivalent to:
 *
 *   gst-launch-1.0 rtpbin name=rtpbin
 *     filesrc location=$MP3 ! decodebin ! audioconvert !
 *     rtpL16pay mtu=$MTU ! rtpbin.send_rtp_sink_1
 *     rtpbin.send_rtp_src_1 !  udpsink port=5000 host=$IP_ADDR
 *     rtpbin.send_rtcp_src_1 ! udpsink port=5001 host=$IP_ADDR
 *                                      sync=false async=false 
 *     udpsrc port=5003 ! rtpbin.recv_rtcp_sink_0
 * 
 *   where:
 *     MP3: name of an MP3 file
 *     MTU: payload size (payload + 12 bytes for RTP header)
 *     IP_ADDR: IP mcast address (e.g "230.0.8.25")
 */
int psa_gst_send_init(char *snd_file, int mtu, char *ip_addr, int latency)
{
        GstElement *source, *decoder, *convert, *pay;
        GstElement *rtpsink, *rtcpsink, *rtcpsrc;
        GstPad *srcpad, *sinkpad;
        GstBus *bus;
        RTPSession *rtpsession;
        GstStateChangeReturn ret;

        g_print("psa_gst_send_init() start\n");

	// l'existence du fichier doit être faite avant l'appel...
	if (snd_file == NULL) {
                g_printerr("psa_gst_send_init(): snd_file error (NULL)\n");
                return 0;
        }

        talker_loop = g_main_loop_new(NULL, FALSE);

        send_pipeline = gst_pipeline_new("psa-lab");
        if (send_pipeline == NULL) {
                g_printerr("talker: pipeline could not be created.\n");
                return -1;
        }

        bus = gst_pipeline_get_bus(GST_PIPELINE(send_pipeline));
        gst_bus_add_watch (bus, bus_call_cb, talker_loop);
        gst_object_unref(bus);

        source = gst_element_factory_make("filesrc", "source");
        if (source == NULL) {
                g_printerr("talker: filesrc could not be created.\n");
                return -1;
        }
        g_object_set(source, "location", snd_file, NULL);
        
        decoder = gst_element_factory_make("decodebin", "decoder");
        if (decoder == NULL) {
                g_printerr("talker: decodebin could not be created.\n");
                return -1;
        }
        
        convert = gst_element_factory_make("audioconvert", "convert");
        if (convert == NULL) {
                g_printerr("talker: audioconvert could not be created.\n");
                return -1;
        }
        
        pay = gst_element_factory_make("rtpL16pay", "pay");
        if (pay == NULL) {
                g_printerr("talker: rtpL16pay could not be created.\n");
                return -1;
        }
        g_object_set(pay, "mtu", mtu, NULL);

        /* Build the pipeline */
        gst_bin_add_many(GST_BIN(send_pipeline), source, decoder,
                         convert, pay, NULL);

        if (gst_element_link(source, decoder) != TRUE) {
                g_printerr("talker: source and decoder could not be linked.\n");
                gst_object_unref(send_pipeline);
                return -1;
        }

        if (gst_element_link(convert, pay) != TRUE) {
                g_printerr("talker: convert and pay could not be linked.\n");
                gst_object_unref(send_pipeline);
                return -1;
        }

        /* decodebin ne peut pas être linké statiquement avec audioconvert
         * il faut le faire dynamiquement, en lien avec le signal pad-added,
         * voir la callback talker_new_pad_cb */
        g_signal_connect (decoder, "pad-added", G_CALLBACK (talker_new_pad_cb),
                          convert);
        
	rtpsink = gst_element_factory_make("udpsink", "rtpsink");
        if (rtpsink == NULL) {
                g_printerr("talker: udpsink rtp could not be created.\n");
                return -1;
        }
                
        rtcpsink = gst_element_factory_make("udpsink", "rtcpsink");
        if (rtpsink == NULL) {
                g_printerr("talker: udpsink rtcp could not be created.\n");
                return -1;
        }
                
        rtcpsrc = gst_element_factory_make("udpsrc", "rtcpsrc");
        if (rtpsink == NULL) {
                g_printerr("talker: udpsrc rtcp could not be created.\n");
                return -1;
        }
                
        send_rtpbin = gst_element_factory_make("rtpbin", "rtpbin");
        if (send_rtpbin == NULL) {
                g_printerr("talker: send_rtpbin could not be created.\n");
                return -1;
        }
        
        gst_bin_add(GST_BIN(send_pipeline), send_rtpbin);
        g_object_set(send_rtpbin, "latency", latency/1000000,
                                  "ntp-time-source", NTP_SOURCE_TIME,
				  NULL);
        
        g_object_set(rtpsink, "port", 5000, "host", ip_addr,
                     "async", FALSE, "sync", TRUE, NULL);
        g_object_set(rtcpsink, "port", 5001, "host", ip_addr,
                     "async", FALSE, "sync", FALSE, NULL);
        g_object_set(rtcpsrc, "port", 5003, NULL);

        gst_bin_add_many(GST_BIN(send_pipeline), rtpsink, rtcpsink,
                         rtcpsrc, NULL);

        /* now link all to the rtpbin, start by getting an RTP
         * sinkpad for session 0 */
        sinkpad = gst_element_get_request_pad(send_rtpbin, "send_rtp_sink_1");
        srcpad = gst_element_get_static_pad(pay, "src");
        if (gst_pad_link(srcpad, sinkpad) != GST_PAD_LINK_OK) {
                g_error("talker: failed to link rtpL16pay to rtpbin");
                return -1;
        }
        gst_object_unref(srcpad);
        
	/* get the RTP srcpad that was created when we requested
         * the sinkpad above and link it to the rtpsink sinkpad */
	srcpad = gst_element_get_static_pad (send_rtpbin, "send_rtp_src_1");
	sinkpad = gst_element_get_static_pad (rtpsink, "sink");
	if (gst_pad_link (srcpad, sinkpad) != GST_PAD_LINK_OK) {
                g_error ("Failed to link rtpbin to rtpsink");
                return -1;
        }
	gst_object_unref (srcpad);
	gst_object_unref (sinkpad);
	
        /* get an RTCP srcpad for sending RTCP to the receiver */
        srcpad = gst_element_get_request_pad(send_rtpbin, "send_rtcp_src_1");
        sinkpad = gst_element_get_static_pad(rtcpsink, "sink");
        if (gst_pad_link(srcpad, sinkpad) != GST_PAD_LINK_OK) {
                g_error("Failed to link rtpbin to rtcpsink");
                return -1;
        }
        //gst_pad_add_probe(srcpad, GST_PAD_PROBE_TYPE_BUFFER,
        //                (GstPadProbeCallback) rtcp_probe_cb, NULL, NULL);
        gst_object_unref(srcpad);

        /* we also want to receive RTCP, request an RTCP sinkpad
         * for session 0 and link it to the srcpad of the udpsrc for RTCP */
        srcpad = gst_element_get_static_pad(rtcpsrc, "src");
        sinkpad = gst_element_get_request_pad(send_rtpbin, "recv_rtcp_sink_0");
        if (gst_pad_link(srcpad, sinkpad) != GST_PAD_LINK_OK) {
                g_error("Failed to link rtcpsrc to rtpbin");
                return -1;
        }
        //gst_pad_add_probe(srcpad, GST_PAD_PROBE_TYPE_BUFFER,
        //                (GstPadProbeCallback) rtcp_probe_cb, NULL, NULL);
        gst_object_unref(srcpad);

        g_signal_emit_by_name(send_rtpbin, "get-internal-session", 0,
                              &rtpsession);
        g_signal_connect(rtpsession, "on-sending-rtcp",
                         G_CALLBACK(on_sending_rtcp_cb), NULL);
        g_signal_connect(rtpsession, "on-receiving-rtcp",
                         G_CALLBACK(on_receiving_rtcp_cb), NULL);

        ret = gst_element_set_state(send_pipeline, GST_STATE_READY);
        if (ret == GST_STATE_CHANGE_FAILURE) {
                g_printerr("Unable to set ready state of the pipeline\n");
                gst_object_unref(send_pipeline);
                return -1;
        }

        g_print("psa_gst_send_init() end\n");
        return 1;
}

int psa_gst_send_play(void)
{
        GstStateChangeReturn ret;
        
        /* Start playing */
        ret = gst_element_set_state(send_pipeline, GST_STATE_PLAYING);
        if (ret == GST_STATE_CHANGE_FAILURE) {
                g_printerr("Unable to set PLAY state of the pipeline\n");
                gst_object_unref(send_pipeline);
                return -1;
        }
        
        g_print("Start playing (%s)...\n",
                   (ret == GST_STATE_CHANGE_SUCCESS)? "SUCCESS"
                   : ((ret == GST_STATE_CHANGE_ASYNC)? "ASYNC": "NO_PREROLL"));

        /* we need to run a GLib main loop to get the messages */
        g_main_loop_run(talker_loop);
        
        g_print("psa_gst_send_play() end\n");
        return 1;
}

int psa_gst_send_pause(void)
{
        GstStateChangeReturn ret;
        /* Start pausing */
        ret = gst_element_set_state(send_pipeline, GST_STATE_PAUSED);
        if (ret == GST_STATE_CHANGE_FAILURE) {
                g_printerr("Unable to set PAUSE state of the pipeline\n");
                gst_object_unref(send_pipeline);
                return -1;
        }
        g_print("PAUSE (%s)...\n", (ret == 1)? "SUCCESS":
                   ((ret == 2)? "ASYNC" : "NO_PREROLL"));

        g_print("psa_gst_send_pause() end\n");
        return 1;
}

int psa_gst_send_close(void) {
        gst_element_set_state(send_pipeline, GST_STATE_NULL);
        gst_object_unref(send_pipeline);
        send_pipeline = NULL;
        send_rtpbin =  NULL;
        return 1;
}

/*
 *
 * listener side
 *
 *
 */


/* 
 * listener_pad_added_cb will be called when rtpbin has validated a payload
 * that we can depayload
 */
static void listener_pad_added_cb(G_GNUC_UNUSED GstElement *not_used,
	                          GstPad *new_pad, GstElement *depay)
{
        GstPad *sinkpad;

        g_print("new payload on pad: %s\n", GST_PAD_NAME(new_pad));
        sinkpad = gst_element_get_static_pad(depay, "sink");
	if (sinkpad == NULL) {
		g_error("Strange, depayloader not accessible...");
	}

        if (gst_pad_link(new_pad, sinkpad) != GST_PAD_LINK_OK) {
		g_error("pad link failed, talker respawn?");
	}
        gst_object_unref(sinkpad);
}

/*
 * psa_gst_recv_init builds a pipeline equivalent to:
 *
 *   gst-launch-1.0 rtpbin name=rtpbin
 *     udpsrc port=5000 caps=$CAPS ! rtpbin.recv_rtp_sink_0
 *     rtpbin. ! rtpL16depay ! audioconvert ! pulsesink
 *     udpsrc port=5001 caps="application/x-rtcp" ! rtpbin.recv_rtcp_sink_0
 *     rtpbin.send_rtcp_src_0 ! udpsink port=5003 host=$IP_ADDR
 *                                      sync=false async=false
 *   where:
 *     IP_ADDR: talker IP address
 *     CAPS: see MP3_CAPS
 *
 *   GStreamer should be initialized before the call
 */
int psa_gst_recv_init(char *ip_addr, char *caps) {
        GstElement *depay, *convert, *sink;
        GstElement *rtpsrc, *rtcpsrc, *rtcpsink;
        GstCaps *the_caps;
        GstPad *srcpad, *sinkpad;
        GstBus *bus;
        RTPSession *rtpsession;
        GstStateChangeReturn ret;
        
        g_print("psa_gst_recv_init() start\n");

        listener_loop = g_main_loop_new(NULL, FALSE);

        recv_pipeline = gst_pipeline_new("psa-lab");
        if (recv_pipeline == NULL) {
                g_printerr("listener: pipeline could not be created.\n");
                return -1;
        }
                
        bus = gst_pipeline_get_bus(GST_PIPELINE(recv_pipeline));
        gst_bus_add_watch(bus, bus_call_cb, listener_loop);
        gst_object_unref(bus);

        /* Create the elements */
        rtpsrc = gst_element_factory_make("udpsrc", "rtpsrc");
        if (rtpsrc == NULL) {
                g_printerr("listener: udpsrc rtp could not be created.\n");
                return -1;
        }
        g_object_set(rtpsrc, "port", 5000, NULL);
        if (caps == NULL)
                the_caps = gst_caps_from_string(MP3_CAPS);
        else
                the_caps = gst_caps_from_string(caps);

        g_object_set(rtpsrc, "caps", the_caps, NULL);
        gst_caps_unref(the_caps);
        
        rtcpsrc = gst_element_factory_make("udpsrc", "rtcpsrc");
        if (rtcpsrc == NULL) {
                g_printerr("listener: udpsrc rtcp could not be created.\n");
                return -1;
        }
        g_object_set(rtcpsrc, "port", 5001, NULL);
	the_caps = gst_caps_from_string("application/x-rtcp");
        g_object_set(rtcpsrc, "caps", the_caps, NULL);
        gst_caps_unref(the_caps);

        rtcpsink = gst_element_factory_make("udpsink", "rtcpsink");
        if (rtcpsink == NULL) {
                g_printerr("listener: udpsink rtcp could not be created.\n");
                return -1;
        }
        g_object_set(rtcpsink, "port", 5003, "host", ip_addr,
                               "async", FALSE, "sync", FALSE, NULL);

        gst_bin_add_many(GST_BIN(recv_pipeline), rtpsrc, rtcpsrc,
                         rtcpsink, NULL);
        
        depay = gst_element_factory_make("rtpL16depay", "depay");
        if (depay == NULL) {
                g_printerr("listener: rtpL16depay could not be created.\n");
                return -1;
        }
        
        convert = gst_element_factory_make("audioconvert", "convert");
        if (convert == NULL) {
                g_printerr("listener: audioconvert could not be created.\n");
                return -1;
        }
        sink = gst_element_factory_make("pulsesink", "sink");
        if (sink == NULL) {
                g_printerr("listener: pulsesink could not be created.\n");
                return -1;
        }

        gst_bin_add_many(GST_BIN(recv_pipeline), depay, convert, sink, NULL);

	if (gst_element_link_many( depay, convert, sink, NULL) != TRUE) {
                g_printerr("listener: elements could not be linked.\n");
                gst_object_unref(recv_pipeline);
                return -1;
        }

        recv_rtpbin = gst_element_factory_make("rtpbin", "rtpbin");
        if (recv_rtpbin == NULL) {
                g_printerr("listener: rtpbin could not be created.\n");
                return -1;
        }
        g_object_set(recv_rtpbin, "ntp-time-source", NTP_SOURCE_TIME, NULL);
        gst_bin_add(GST_BIN(recv_pipeline), recv_rtpbin);

        /* now link all to the rtpbin, start by getting an RTP sinkpad
         * for session 0 */
        srcpad = gst_element_get_static_pad(rtpsrc, "src");
        sinkpad = gst_element_get_request_pad(recv_rtpbin, "recv_rtp_sink_0");
        if (gst_pad_link(srcpad, sinkpad) != GST_PAD_LINK_OK) {
		g_error("failed to link src to recv_rtp_sink_0");
		return -1;
	}
        gst_object_unref(srcpad);

        /* get an RTCP sinkpad in session 0 */
        srcpad = gst_element_get_static_pad(rtcpsrc, "src");
        sinkpad = gst_element_get_request_pad(recv_rtpbin, "recv_rtcp_sink_0");
        if (gst_pad_link(srcpad, sinkpad) != GST_PAD_LINK_OK) {
		g_error("failed to link src to recv_rtcp_sink_0");
		return -1;
	}
        //gst_pad_add_probe(sinkpad, GST_PAD_PROBE_TYPE_BUFFER,
        //                (GstPadProbeCallback) rtcp_probe_cb, NULL, NULL);
        gst_object_unref(srcpad);
        gst_object_unref(sinkpad);
        
        /* get an RTCP srcpad for sending RTCP back to the sender */
        srcpad = gst_element_get_request_pad(recv_rtpbin, "send_rtcp_src_0");
        sinkpad = gst_element_get_static_pad(rtcpsink, "sink");
        if (gst_pad_link(srcpad, sinkpad) != GST_PAD_LINK_OK) {
		g_error("failed to link send_rtcp_src_0 to sink");
		return -1;
	}
        //gst_pad_add_probe(sinkpad, GST_PAD_PROBE_TYPE_BUFFER,
        //                (GstPadProbeCallback) rtcp_probe_cb, NULL, NULL);
        gst_object_unref(sinkpad);

        /* the RTP pad that we have to connect to the depayloader will be
         * created dynamically so we connect to the pad-added signal,
         * pass the depayloader as user_data so that we can link to it. */
        g_signal_connect(recv_rtpbin, "pad-added",
                         G_CALLBACK(listener_pad_added_cb), depay);

        /* give some stats when we receive RTCP
	   g_signal_connect(recv_rtpbin, "on-ssrc-active",
	   G_CALLBACK(listener_on_ssrc_active_cb), depay);
        */

        g_signal_emit_by_name(recv_rtpbin, "get-internal-session", 0,
                              &rtpsession);
        g_signal_connect(rtpsession, "on-sending-rtcp",
                         G_CALLBACK(on_sending_rtcp_cb), NULL);
        g_signal_connect(rtpsession, "on-receiving-rtcp",
                         G_CALLBACK(on_receiving_rtcp_cb), NULL);
        
        ret = gst_element_set_state(recv_pipeline, GST_STATE_READY);
        if (ret == GST_STATE_CHANGE_FAILURE) {
                g_printerr("Unable to set ready state of the pipeline\n");
                gst_object_unref(recv_pipeline);
                return -1;
        }
        
        g_print("psa_gst_recv_init() end\n");
        return 1;
}

int psa_gst_recv_play(void)
{
        GstStateChangeReturn ret;
        
        /* Start playing */
        ret = gst_element_set_state(recv_pipeline, GST_STATE_PLAYING);
        if (ret == GST_STATE_CHANGE_FAILURE) {
                g_printerr("Unable to set PLAY state of the pipeline\n");
                gst_object_unref(recv_pipeline);
                return -1;
        }
        g_print("Start playing...\n");

        /* we need to run a GLib main loop to get the messages */
        g_main_loop_run(listener_loop);

        g_print("psa_gst_recv_play() end\n");
        
	return 1;
}

int psa_gst_recv_pause(void)
{
        GstStateChangeReturn ret;
        
        /* Start pause */
        ret = gst_element_set_state(recv_pipeline, GST_STATE_PAUSED);
        if (ret == GST_STATE_CHANGE_FAILURE) {
                g_printerr("Unable to set PAUSE state of the pipeline\n");
                gst_object_unref(recv_pipeline);
                return -1;
        }
        g_print("PAUSE...\n");

        g_print("psa_gst_recv_pause() end\n");
        
	return 1;
}

int psa_gst_recv_close(void)
{
        gst_element_set_state(recv_pipeline, GST_STATE_NULL);
        gst_object_unref(recv_pipeline);
        recv_pipeline = NULL;
        recv_rtpbin = NULL;
        return 1;
}
